# odate.rb - an imitation of Yasuo Ohba's date class: coded by tadf 1998-2000
# $Id: odate.rb,v 1.8 2000-07-16 10:23:59+09 tadf Exp $

class Date

  include Comparable

  IDENT = 1

  Monthtab = { 'jan'=>1, 'feb'=>2, 'mar'=>3, 'apr'=>4,  'may'=>5,  'jun'=>6,
               'jul'=>7, 'aug'=>8, 'sep'=>9, 'oct'=>10, 'nov'=>11, 'dec'=>12 }

  Monthtag = %w(January February March April May June July
                August September October November December)

  Weektag = %w(Sunday Monday Tuesday Wednesday Thursday Friday Saturday)

  def Date.period(y, m, d)
    if m <= 2
      y -= 1
      m += 12
    end
    a = (y / 100.0).floor
    b = 2 - a + (a / 4.0).floor
    jd = (365.25 * (y + 4716)).floor +
      (30.6001 * (m + 1)).floor +
      d + b - 1524
    unless jd >= 2361222
      jd -= b
    end
    jd - 1721423
  end

  def Date.civil(dt)
    case dt
    when Time; return dt.year, dt.mon, dt.mday
    when Date; return dt.year, dt.month, dt.day
    end
    jd = dt + 1721423
    unless jd >= 2361222
      a = jd
    else
      x = ((jd - 1867216.25) / 36524.25).floor
      a = jd + 1 + x - (x / 4.0).floor
    end
    b = a + 1524
    c = ((b - 122.1) / 365.25).floor
    d = (365.25 * c).floor
    e = ((b - d) / 30.6001).floor
    dom = b - d - (30.6001 * e).floor
    if e <= 13
      m = e - 1
      y = c - 4716
    else
      m = e - 13
      y = c - 4715
    end
    return y, m, dom
  end

  def Date.at(dt) Date.new(*Date.civil(dt)) end

  def initialize(y=1, m=1, d=1)
    if y.kind_of? String
      if /^(\d{4})-?(?:(\d{2})-?(\d{2})?)?/no =~ y
	y = $1.to_i
	m = $2.to_i if $2
	d = $3.to_i if $3
      else
	require 'parsedate'
	y, m, d = ParseDate.parsedate(y)
      end
    else
      if m.kind_of? String
	if /^([a-z]{3})/ino =~ m
	  m = Monthtab[$1.downcase]
	end
      end
      y, m, d = y.to_i, m.to_i, d.to_i
    end
    @period = Date.period(@year = y, @month = m, @day = d)
    unless [y, m, d] == Date.civil(@period)
      raise ArgumentError, 'invalid date'
    end
  end

  def period() @period end
  def jd() @period + 1721423 end
  def mjd() jd - 2400000.5 end
  def year() @year end
  def month() @month end
  def day() @day end
  def day_of_week() (@period + 5) % 7 end
  def leapyear? () (Date.new(@year, 3, 1) - 1).day == 29 end

  def + (n)
    case n
    when Numeric; return Date.at(@period + n)
    end
    raise TypeError, 'expected numeric'
  end

  def - (x)
    case x
    when Numeric; return Date.at(@period - x)
    when Date;    return @period - x.period
    end
    raise TypeError, 'expected numeric or date'
  end

  def <=> (other)
    case other
    when Numeric; return @period <=> other
    when Date;    return @period <=> other.period
    end
    raise TypeError, 'expected numeric or date'
  end

  def eql? (other) Date === other and self == other end
  def hash() @period end

  def name_of_month() Monthtag[@month - 1] end
  def name_of_week() Weektag[day_of_week] end

  def to_s
    format('%.3s, %.3s %2d %4d', name_of_week, name_of_month, @day, @year)
  end

  def inspect() to_s end

end
